#include <math.h>
#include "KnobP.h"

/* defines for mathmatic functions used in drawing */
#define  RADIANS(x)  (M_PI * 2.0 * (x) / 360.0)
#define  DEGREES(x)  ((x) / (M_PI * 2.0) * 360.0)
#define  MIN_ANGLE   225.0
#define  MAX_ANGLE   345.0
#define  NUM_SEGS    24
#define  MIN(a,b)    (((a) < (b)) ? (a) :  (b))

#define MIN_KNOB_DIAMETER		3
#define DEFAULT_DIAMETER		100
#define DEFAULT_KNOB_MARGIN		5
#define DEFAULT_KNOB_MARGIN_PERCENTAGE	50

/********    Static Function Declarations    ********/

/* core class methods */
static void ClassPartInitialize( WidgetClass widget_class) ;
static void Initialize( Widget request, Widget new_w, ArgList args,
                        Cardinal *num_args) ;
static void Destroy( Widget w) ;
static void Resize( Widget w) ;
static void Redisplay( Widget w, XEvent *event, Region region) ;
static Boolean SetValues( Widget old_w, Widget request, Widget new_w,
                          ArgList args, Cardinal *num_args) ;
static void SetValuesAlmost( Widget old_w, Widget new_w,
                            XtWidgetGeometry *request, XtWidgetGeometry *reply);
static XtGeometryResult QueryGeometry( Widget w, XtWidgetGeometry *request, 
                             	       XtWidgetGeometry *reply) ;

/* primitive class methods */
static void HighlightKnob( Widget w) ;
static void UnhighlightKnob( Widget w) ;

/* knob class methods */
static void GetDiameters( Widget w, unsigned int *diameter,
                         unsigned int *inner_diameter) ;
static void CreateSegments( Widget w) ;
static void DrawIndicator( Widget w) ;

/* resource callprocs */
static void DefaultKnobColor( Widget, int, XrmValue *);

/* action procedures */
static void TurnKnob( XtPointer closure, XtIntervalId *id) ;
static void DrawKnob( Widget w) ;
static void TurnLeft( Widget w, XEvent *event, char **params,
                      Cardinal *num_params) ;
static void TurnRight( Widget w, XEvent *event, char **params,
                       Cardinal *num_params) ;
static void ToggleLeft( Widget w, XEvent *event, char **params,
                        Cardinal *num_params) ;
static void ToggleRight( Widget w, XEvent *event, char **params,
                         Cardinal *num_params) ;
static void ReleaseKnob( Widget w, XEvent *event, char **params,
                         Cardinal *num_params) ;


/* Resource list initialization - sets up resource names and values
 * for each resource of this widget.
 */
static XtResource resources[] = {
	{
	  XsmNvalueChangedCallback, XmCCallback, XmRCallback,
	  sizeof(XtCallbackList),
	  XtOffset (XsmKnobWidget, knob.value_changed_callback),
	  XmRImmediate, NULL
        },

	{
	  XsmNvalue, XsmCValue, XmRInt, sizeof(int),
	  XtOffset (XsmKnobWidget, knob.value),
	  XmRImmediate, (XtPointer) 0
        },
	  
	{
	  XsmNmaxValue, XsmCMaxValue, XmRInt, sizeof(int),
	  XtOffset (XsmKnobWidget, knob.max_val),
	  XmRImmediate, (XtPointer) 100
        },
	  
	{
	  XsmNminValue, XsmCMinValue, XmRInt, sizeof(int),
	  XtOffset (XsmKnobWidget, knob.min_val),
	  XmRImmediate, (XtPointer) 0
        },
	  
        {
          XsmNturnDelay, XsmCTurnDelay, XmRInt, sizeof(int),
          XtOffset(XsmKnobWidget, knob.turn_delay),
          XmRImmediate, (XtPointer) 50
        },

	{
	  XsmNknobColor, XmCBackground, XmRPixel, sizeof (Pixel),
	  XtOffset (XsmKnobWidget, knob.knob_color),
	  XmRCallProc, (XtPointer) DefaultKnobColor
        },
	  
	{
	  XsmNindicatorColor, XsmCIndicatorColor, XmRPixel, sizeof (Pixel),
	  XtOffset (XsmKnobWidget, knob.indicator_color),
	  XmRCallProc, (XtPointer) _XmForegroundColorDefault
        },

	{
	  XsmNmarginWidth, XmCMarginWidth, XmRHorizontalDimension,
	  sizeof (Dimension),
	  XtOffset (XsmKnobWidget, knob.margin_width),
	  XmRImmediate, (XtPointer) DEFAULT_KNOB_MARGIN
        },
	  
	{
	  XsmNmarginHeight, XmCMarginHeight, XmRVerticalDimension,
	  sizeof (Dimension),
	  XtOffset (XsmKnobWidget, knob.margin_height),
	  XmRImmediate, (XtPointer) DEFAULT_KNOB_MARGIN
        },
	  
	{
	  XsmNknobMargin, XsmCKnobMargin, XmRDimension, sizeof (Dimension),
	  XtOffset (XsmKnobWidget, knob.knob_margin),
	  XmRImmediate, (XtPointer) DEFAULT_KNOB_MARGIN_PERCENTAGE
        },
};
	  
/* Synthetic resource list initialization - sets up resource procedures
 * to get called when setting or getting the various resources.  These
 * synthetic resources are used in resolution independance.
 */
static XmSyntheticResource syn_resources[] = {
        {
          XsmNmarginWidth,
          sizeof (Dimension),
          XtOffset( XsmKnobWidget, knob.margin_width),
          _XmFromHorizontalPixels,
          _XmToHorizontalPixels
        },

        {
          XsmNmarginHeight,
          sizeof (Dimension),
          XtOffset( XsmKnobWidget, knob.margin_height),
          _XmFromVerticalPixels,
          _XmToVerticalPixels
        }
};


/* Action table - where procedures are mapped to strings */
static XtActionsRec knob_actions[] = {
    {"turn-left",		TurnLeft},
    {"turn-right",		TurnRight},
    {"toggle-left",		ToggleLeft},
    {"toggle-right",		ToggleRight},
    {"release-knob",		ReleaseKnob},
};

/* Translation table - where events are mapped to action names */
char defaultTranslations[] = "\
<Key>greater:	toggle-left()\n\
<Key>less:	toggle-right()\n\
<Key>:		release-knob()\n\
<Btn1Down>:	turn-left()\n\
<Btn3Down>:	turn-right()\n\
<Btn1Up>:	release-knob()\n\
<Btn3Up>:	release-knob()\n\
<FocusIn>:	PrimitiveFocusIn()\n\
<FocusOut>:	PrimitiveFocusOut()\n\
<EnterWindow>:	PrimitiveEnter()\n\
<LeaveWindow>:	PrimitiveLeave()";


/* Widget class record initialization - Here we initialize each field of the
 * core, primitive, and knob class parts.
 */
externaldef (xsmknobclassrec) XsmKnobClassRec xsmKnobClassRec = {
   {
      (WidgetClass) &xmPrimitiveClassRec,       /* superclass         */
      "XsmKnob",                            	/* class_name         */
      sizeof(XsmKnobRec),                   	/* widget_size        */
      NULL,                          		/* class_initialize   */
      ClassPartInitialize,             		/* class_part_initiali*/
      FALSE,                                    /* class_inited       */
      Initialize,                               /* initialize         */
      NULL,                                     /* initialize_hook    */
      XtInheritRealize,                         /* realize            */
      knob_actions,                             /* actions            */
      XtNumber(knob_actions),                   /* num_actions        */
      resources,                                /* resources          */
      XtNumber(resources),                      /* num_resources      */
      NULLQUARK,                                /* xrm_class          */
      TRUE,                                     /* compress_motion    */
      XtExposeCompressMaximal,                  /* compress_exposure  */
      TRUE,                                     /* compress_enterleave*/
      FALSE,                                    /* visible_interest   */
      Destroy,                                  /* destroy            */
      Resize,                                   /* resize             */
      Redisplay,                      		/* expose             */
      SetValues,                                /* set_values         */
      NULL,                                     /* set_values_hook    */
      SetValuesAlmost,                 		/* set_values_almost  */
      NULL,                                     /* get_values_hook    */
      NULL,                                     /* accept_focus       */
      XtVersion,                                /* version            */
      NULL,                                     /* callback_private   */
      defaultTranslations,                      /* tm_table           */
      QueryGeometry,                   		/* query_geometry     */
      XtInheritDisplayAccelerator,              /* display accel      */
      NULL,                                     /* extension          */
   },

   {                                            /* Xmprimitive        */
      HighlightKnob,                            /* border_highlight   */
      UnhighlightKnob,                          /* border_unhighlight */
      XtInheritTranslations,                    /* translations       */
      NULL,                                     /* arm_and_activate   */
      syn_resources,                   		/* syn resources      */
      XtNumber(syn_resources),			/* num syn_resources  */
      NULL,     				/* extension          */
   },

   {                                            /* knob class 		*/
      GetDiameters,				/* get_diameters	*/
      CreateSegments,				/* create_segments	*/
      DrawIndicator,				/* draw_indicator	*/
      TurnKnob,					/* turn			*/
      DrawKnob,					/* draw			*/
      NULL,                                     /* extension            */
   }
};

externaldef(xsmknobwidgetclass) WidgetClass xsmKnobWidgetClass =
                                         (WidgetClass) &xsmKnobClassRec;

/****************************************************************
 *
 * Core Class Methods
 *
 ****************************************************************/


/************************************************************************
 *
 *  ClassPartInitialize() -
 *      The initialization routine for class parts.  It is called for the
 *	initialization of this class and each time a new subclass is created
 *	for this widget class.  It is called only once per subclass.
 *
 ************************************************************************/
static void ClassPartInitialize(WidgetClass wc)
{
    XsmKnobWidgetClass kc = (XsmKnobWidgetClass) wc;
    XsmKnobWidgetClass sc = (XsmKnobWidgetClass) wc->core_class.superclass;

   /* assign procedures to the classes pointers that inherit these procedures */

    if (kc->knob_class.get_diameters == XsmInheritGetDiameters)
       kc->knob_class.get_diameters = sc->knob_class.get_diameters;

    if (kc->knob_class.create_segments == XsmInheritCreateSegments)
       kc->knob_class.create_segments = sc->knob_class.create_segments;

    if (kc->knob_class.draw_indicator == XsmInheritDrawIndicator)
       kc->knob_class.draw_indicator = sc->knob_class.draw_indicator;

    if (kc->knob_class.turn == XsmInheritTurn)
       kc->knob_class.turn = sc->knob_class.turn;

    if (kc->knob_class.draw == XsmInheritDraw)
       kc->knob_class.draw = sc->knob_class.draw;

}



/************************************************************************
 *
 *  Initialize() -
 *      The main widget instance initialization routine.  Verify resource
 *  values and initialize widget instance fields.
 *
 ************************************************************************/

/* ARGSUSED */
static void Initialize( Widget request, Widget new_w, ArgList args,
                        Cardinal *num_args )
{
    XsmKnobWidget req_kw = (XsmKnobWidget) request;
    XsmKnobWidget new_kw = (XsmKnobWidget) new_w;
    XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(new_w);
    Dimension min_width, min_height;
    XGCValues values;
    unsigned long valuemask;

   /* Verify resource values */

   /* verify that the maximum value is not less than or equal to zero */
    if (new_kw->knob.max_val <= 0) {
      XtWarning("MaxValue is must be greater than 0, Defaulting MaxValue to 1");
      new_kw->knob.max_val = 1;
    }

   /*
    * verify that the maximum value is not less than the minimum value
    */
    if (new_kw->knob.max_val < new_kw->knob.min_val) {
       XtWarning("MaxValue is less than the MinValue, Defaulting to MinValue to 0");
       new_kw->knob.min_val = 0;
    }

   /* verify that the value is not greater than the maximum value */
    if (new_kw->knob.value > new_kw->knob.max_val) {
       XtWarning("Value is greater than MaxValue, Defaulting to MaxValue");
       new_kw->knob.value = new_kw->knob.max_val;
    }

   /* verify that the value is not less than the minimum value */
    if (new_kw->knob.value < new_kw->knob.min_val) {
       XtWarning("Value is less than MinValue, Defaulting to MinValue");
       new_kw->knob.value = new_kw->knob.min_val;
    }

   /* verify that turn rate is not less than 1 */
    if (new_kw->knob.turn_delay < 1) {
       XtWarning("Turn Delay is less than 1, Defaulting to 1");
       new_kw->knob.turn_delay = 1;
    }

   /*
    * Knob margin is a percentage of the total knob,
    * thus it must be between 0 and 100.  Since this resource is
    * typed as a dimension (unsigned short), a negative value
    * becomes a large positive, thus this also checks for < 0/
    */

    if (new_kw->knob.knob_margin > 100)
    {
       XtWarning("Knob Margin must be between 0 and 100");
       new_kw->knob.knob_margin = DEFAULT_KNOB_MARGIN_PERCENTAGE;
    }
       

   /*
    * initialize the minimum required dimensions for the knob
    */

    min_width = MIN_KNOB_DIAMETER +
	        (2 * (new_kw->knob.margin_width +
	              new_kw->primitive.highlight_thickness));
    min_height = MIN_KNOB_DIAMETER +
                 (2 * (new_kw->knob.margin_width +
		       new_kw->primitive.highlight_thickness));
		
   /*
    * if an initial width and height weren't set, initialize
    * a default dimensions, otherwise verify that the dimensions
    * that were set are at least the minimum dimension
    */

    if (req_kw->core.width == 0)
       new_kw->core.width = DEFAULT_DIAMETER +
                            (2 * (new_kw->knob.margin_width +
			          new_kw->primitive.highlight_thickness));
    else
       if (new_kw->core.width < min_width)
          new_kw->core.width = min_width;

    if (req_kw->core.height == 0)
       new_kw->core.height = DEFAULT_DIAMETER +
                            (2 * (new_kw->knob.margin_height +
			          new_kw->primitive.highlight_thickness));
    else
       if (new_kw->core.height < min_width)
          new_kw->core.height = min_height;



   /* save original calculated width and height for use in query geometry */
       new_kw->knob.orig_width = new_kw->core.width;
       new_kw->knob.orig_height = new_kw->core.height;

   /* Initialize internal variables */

    new_kw->knob.num_segments = NUM_SEGS;
    new_kw->knob.segments = (XSegment *) XtMalloc(sizeof(XSegment) *
				    (new_kw->knob.num_segments + 1));

    new_kw->knob.move_clockwise = True;
    new_kw->knob.turning = False;
    new_kw->knob.timer_id = NULL;
    new_kw->knob.angle_offset = 0;

   /* Create initial segments */

    (*kc->knob_class.create_segments)(new_w);

   /* Set up graphic context for use in drawing the knob */

    valuemask = (GCForeground | GCBackground);
    values.foreground = new_kw->primitive.foreground;
    values.background = new_kw->knob.knob_color;
    new_kw->knob.gc = XtGetGC(new_w, valuemask, &values); 
}



/************************************************************************
 *
 *  Destroy() -
 *      Clean up allocated resources when the widget is destroyed.
 *	Free any allocated data.
 *
 ************************************************************************/
static void Destroy( Widget w )
{
    XsmKnobWidget kw = (XsmKnobWidget) w;
   
   /* Free graphics contexts */
    XtReleaseGC( w, kw->knob.gc);
    
   /* Free allocated data */
    XtFree((char *) kw->knob.segments);

   /* Remove any outstanding timeouts */
    if (kw->knob.timer_id) XtRemoveTimeOut(kw->knob.timer_id);

   /* Remove all callbacks */
    XtRemoveAllCallbacks (w, XsmNvalueChangedCallback);
}


/************************************************************************
 *
 *  Resize() -
 *      Redraw according to when size changes.
 *
 ************************************************************************/
static void Resize( Widget w )
{
    XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);

   /* Recreate any segments based on the new size */
    (*kc->knob_class.create_segments)(w);
}



/************************************************************************
 *
 *  Redisplay() -
 *      Redraw region that go exposed.  Simple and less efficient
 *  expose routines simply redraw the whole widget.  This is an
 *  example of a simple expose routine.
 *
 ************************************************************************/
/* ARGSUSED */
static void Redisplay( Widget w, XEvent *event, Region region)
{
   XsmKnobWidget kw = (XsmKnobWidget) w;
   XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);

  /* Use the class pointer to redraw the whole knob */
   (*kc->knob_class.draw)(w);

  /* Use the class pointers to highlight or unhighlihht the knob */
   if (kw->primitive.highlighted)
      (*kc->primitive_class.border_highlight)(w);
   else
      (*kc->primitive_class.border_unhighlight)(w);
}



/************************************************************************
 *
 *  SetValues() -
 *      The procedure is used to verify changes to the resource values.
 *	If the resource changes effect the visuals of the widget, the
 *	return value of the procedure should be set to True indicating
 *	that the expose procedure will be called to redraw the widget.
 *	Any changes to the widget's geometry in this procedure will
 *	generate a geometry request to it's parent.
 *
 ************************************************************************/
/* ARGSUSED */
static Boolean SetValues( Widget old_w, Widget request, Widget new_w,
	                  ArgList args, Cardinal *num_args )
{
    XsmKnobWidget old_kw = (XsmKnobWidget) old_w;
    XsmKnobWidget new_kw = (XsmKnobWidget) new_w;
    XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(new_w);
    Dimension min_width, min_height;
    Boolean redisplay = False; /* flag to determine return value */

   /* verify changes in resource settings */

   /* verify that the maximum value is not less than or equal to zero */
    if (new_kw->knob.max_val <= 0) {
       XtWarning("MaxValue is must be greater than 0, MaxValue set to 1");
       new_kw->knob.max_val = 1;
    }

   /*
    * verify that the minimum value is not greater or equal to 
    * the maximum value
    */
    if (new_kw->knob.min_val >= new_kw->knob.max_val) {
       XtWarning("MinValue is greater than or equal to MaxValue, MinValue set to 0");
       new_kw->knob.min_val = 0;
    }

   /* verify that the value is not greater than the maximum value */
    if ((new_kw->knob.max_val != old_kw->knob.max_val ||
         new_kw->knob.value != old_kw->knob.value) &&
         new_kw->knob.value > new_kw->knob.max_val) {
       XtWarning("Value is greater than MaxValue, Changing Value to MaxValue");
       new_kw->knob.value = new_kw->knob.max_val;
       redisplay = True;
    }

   /* verify that the value is not less than the minimum value */
    if ((new_kw->knob.min_val != old_kw->knob.min_val ||
         new_kw->knob.value != old_kw->knob.value) &&
         new_kw->knob.value < new_kw->knob.min_val) {
       XtWarning("Value is less than MinValue, Changing Value to MinValue");
       new_kw->knob.value = new_kw->knob.min_val;
       redisplay = True;
    }

   /* verify that turn rate is not less than 1 */
    if (new_kw->knob.turn_delay < 1) {
       XtWarning("Turn Delay is less than 1, Defaulting to 1");
       new_kw->knob.turn_delay = 1;
    }

   /*
    * Knob margin is a percentage of the total knob,
    * thus it must be between 0 and 100.  Since this resource is
    * typed as a dimension (unsigned short), a negative value
    * becomes a large positive, thus this also checks for < 0/
    */
    if (new_kw->knob.knob_margin > 100)
    {
       XtWarning("Knob Margin must be between 0 and 100");
       new_kw->knob.knob_margin = old_kw->knob.knob_margin;
    }

	
   /* Don't allow a zero size to set */
    if (new_kw->core.width == 0) new_kw->core.width = old_kw->core.width;
    if (new_kw->core.height == 0) new_kw->core.height = old_kw->core.height;

   /* If the color resources change, the graphic context needs to be updated */
    if (new_kw->primitive.foreground != old_kw->primitive.foreground ||
        new_kw->knob.knob_color != old_kw->knob.knob_color) {
       XGCValues values;
       unsigned long valuemask;

       valuemask = (GCForeground | GCBackground );
       values.foreground = new_kw->primitive.foreground;
       values.background = new_kw->knob.knob_color;
       XtReleaseGC( new_w, new_kw->knob.gc);
       new_kw->knob.gc = XtGetGC(new_w, valuemask, &values); 
       redisplay = True;
    }

    min_width = MIN_KNOB_DIAMETER +
	        (2 * (new_kw->knob.margin_width +
	              new_kw->primitive.highlight_thickness));
    min_height = MIN_KNOB_DIAMETER +
                 (2 * (new_kw->knob.margin_height +
		       new_kw->primitive.highlight_thickness));
   /* Try to retain a minimum height and width */

    if (new_kw->core.width < min_width)
    {
       new_kw->core.width = min_width;
       redisplay = True;
    }

    if (new_kw->core.height < min_height)
    {
       new_kw->core.height = min_height;
       redisplay = True;
    }

   /*
    * If any of the following resource values changed, the knob
    * segments need to be recomputed and the knob need to be redisplayed
    */
    if (new_kw->knob.margin_height != old_kw->knob.margin_height ||
        new_kw->knob.margin_width != old_kw->knob.margin_width ||
        new_kw->knob.knob_margin != old_kw->knob.knob_margin ||
        new_kw->knob.max_val != old_kw->knob.max_val ||
        new_kw->knob.min_val != old_kw->knob.min_val ||
        new_kw->core.height != old_kw->core.height ||
        new_kw->core.width != old_kw->core.width ||
        new_kw->primitive.highlight_thickness != 
	old_kw->primitive.highlight_thickness) {

       (*kc->knob_class.create_segments)(new_w);
       redisplay = True;
    }

    return(redisplay);
}



/************************************************************************
 *
 *  SetValuesAlmost() -
 *      The procedure is used to negotiate a geometry with the parent
 *	if the change in geometry in the SetValues Procedure failed.
 *	Setting the reply->request_mode to zero ends the negotiations.
 *
 ************************************************************************/
/* ARGSUSED */
static void SetValuesAlmost( Widget old_w, Widget new_w,
                             XtWidgetGeometry *request, XtWidgetGeometry *reply)
{
    XsmKnobWidget new_kw = (XsmKnobWidget) new_w;
    XsmKnobWidget old_kw = (XsmKnobWidget) old_w;
    Dimension min_width, min_height;

   /* If the request failed completely, reset old values and return */
    if (reply->request_mode == 0)
    {
       new_kw->knob.margin_width = old_kw->knob.margin_width;
       new_kw->knob.margin_height = old_kw->knob.margin_height;
       request->request_mode = 0;
       return;
    }

   /* all attempts to accomodate size failed */ 
    if (new_kw->knob.margin_width == 0 ||
        new_kw->knob.margin_height == 0)
    {   
       request->request_mode = 0;
       return;
    }

    min_width = MIN_KNOB_DIAMETER +
	        (2 * (new_kw->knob.margin_width +
	              new_kw->primitive.highlight_thickness));
    min_height = MIN_KNOB_DIAMETER +
                 (2 * (new_kw->knob.margin_width +
		       new_kw->primitive.highlight_thickness));
    
   /* reduce margin to accomodate smaller size */
    if ( reply->width < min_width )
    {
       while ( reply->width < min_width )
       {
          new_kw->knob.margin_width--;
          min_width -= 2; /* minus one pixel on each side of knob */
          if (new_kw->knob.margin_width == 0) break;
       }
       request->width = min_width;
    }
    else /* accept larger size */
    {
       request->width = reply->width;
    }

   /* reduce margin to accomodate smaller size */
    if (reply->height < min_height)
    {
       while ( reply->height < min_height )
       {
          new_kw->knob.margin_height--;
          min_height -= 2; /* minus one pixel on each side of knob */
          if (new_kw->knob.margin_height == 0) break;
       }
       request->height = min_height;
    }
    else /* accept larger size */
    {
       request->width = reply->width;
    }
}

/************************************************************************
 *
 *  QueryGeometry() -
 *      The procedure is used to negotiate a geometry with the parent
 *	when the parent is requesting a preferred of the child.
 *	If the values in the request don't match the values in the
 *	reply, we need to return either XtGeometryNo or XtGeoemtryAlmost.
 *	Use XtGeoemtryNo if the values in the request match the current values.
 *	Note: This procedure is only interested in height and width values.
 *	It will accept any changes to x, y, border width and stacking order.
 *
 ************************************************************************/

static XtGeometryResult QueryGeometry( Widget w, XtWidgetGeometry *request, 
                             	       XtWidgetGeometry *reply)
{
    XsmKnobWidget kw = (XsmKnobWidget) w;

    reply->width = kw->knob.orig_width;
    reply->height = kw->knob.orig_height;
    reply->request_mode = CWWidth | CWHeight;

   /* return XtGeometryYes if the request matches the reply width and height */
    if (((request->request_mode & CWWidth) && request->width == reply->width) &&
       ((request->request_mode & CWHeight) && request->height == reply->height))
       return XtGeometryYes;

   /* return XtGeometryNo if reply matches our current width and height */
    if (reply->width == kw->core.width && reply->height == kw->core.height)
       return XtGeometryNo; 
    
   /* return XtGeometryAlmost if one of the reply fields doesn't match the
    * current or request width or height.
    */
    return XtGeometryAlmost; 
}

/****************************************************************
 *
 * Primitive Class Methods
 *
 ****************************************************************/


/************************************************************************
 *
 * HighlightKnob
 *	The is a primitive class procedure for drawing a highlight around
 * the knob.  Knob needs a customized one to draw a circular highlight.
 *
 ************************************************************************/
static void HighlightKnob( Widget w )
{
   XsmKnobWidget kw = (XsmKnobWidget) w;
   XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);
   unsigned int outer_diameter, inner_diameter, diameter;
   XGCValues values;
   unsigned long valuemask;
    
  /* Get the diameter */
   (*kc->knob_class.get_diameters)(w, &outer_diameter, &inner_diameter);

   diameter = outer_diameter + (2 * kw->primitive.highlight_thickness);

  /* Change the graphic context to highlight the knob */
   valuemask = (GCLineWidth | GCForeground);
   values.foreground = kw->primitive.highlight_color;
   values.line_width = kw->primitive.highlight_thickness;
   XChangeGC(XtDisplay(kw), kw->knob.gc, valuemask, &values);

  /* Draw the highlight */
   XDrawArc(XtDisplay(kw), XtWindow(kw), kw->knob.gc, kw->knob.margin_width,
	    kw->knob.margin_height, diameter, diameter, 0, 64*360);

  /* Set flags to indicate that the highlight has been drawn */
   kw->primitive.highlighted = True;
   kw->primitive.highlight_drawn = True;

  /* Change the graphic context back for drawing the knob */
   values.foreground = kw->knob.knob_color;
   values.line_width = 0;
   XChangeGC(XtDisplay(kw), kw->knob.gc, valuemask, &values);
}


/************************************************************************
 *
 * UnhighlightKnob
 *	The is a primitive class procedure for erasing a highlight around
 * the knob.  Knob needs a customized one to erase a circular highlight.
 *
 ************************************************************************/
static void UnhighlightKnob( Widget w )
{
   XsmKnobWidget kw = (XsmKnobWidget) w;
   XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);
   unsigned int outer_diameter, inner_diameter, diameter;
   XGCValues values;
   unsigned long valuemask;
    
  /* Get the diameter */
   (*kc->knob_class.get_diameters)(w, &outer_diameter, &inner_diameter);

   diameter = outer_diameter + (2 * kw->primitive.highlight_thickness);

  /* Change the graphic context to unhighlight */
   valuemask = (GCLineWidth | GCForeground);
   values.foreground = kw->core.background_pixel;
   values.line_width = kw->primitive.highlight_thickness;
   XChangeGC(XtDisplay(kw), kw->knob.gc, valuemask, &values);

  /* Erase the highlight */
   XDrawArc(XtDisplay(kw), XtWindow(kw), kw->knob.gc, kw->knob.margin_width,
	    kw->knob.margin_height, diameter, diameter, 0, 64*360);

   kw->primitive.highlighted = False;
   kw->primitive.highlight_drawn = False;

  /* Change the graphic context back for drawing the knob */
   values.foreground = kw->knob.knob_color;
   values.line_width = 0;
   XChangeGC(XtDisplay(kw), kw->knob.gc, valuemask, &values);

}

/****************************************************************
 *
 * Knob Class Methods
 *
 ****************************************************************/


/************************************************************************
 *
 * GetDiameters
 * 	This knob class procedure is used to determine the inner and out
 * diameter of the knob.  The knob draws 3 circles. The outer most circle
 * is for highlighting. The next circle is for drawing the outer edge of
 * the knob.  The margin_width or the margin_height determine the outer edge
 * of the circle.  The knob_margin determines the inner most circle (being
 * a percentage of the knob itself).
 *
 ************************************************************************/
static void GetDiameters( Widget w, unsigned int *diameter,
			 unsigned int *inner_diameter)
{
    XsmKnobWidget kw = (XsmKnobWidget) w;
    
    if ((Dimension) (kw->core.height - (2 * kw->knob.margin_width)) >=
        (Dimension) (kw->core.width - (2 * kw->knob.margin_height)))
       *diameter = kw->core.width - (2 * (kw->knob.margin_width + 
		                          kw->primitive.highlight_thickness));
    else
       *diameter = kw->core.height - (2 * (kw->knob.margin_height +
		                           kw->primitive.highlight_thickness));

    if (*diameter < MIN_KNOB_DIAMETER) *diameter = MIN_KNOB_DIAMETER;

    *inner_diameter = (*diameter * kw->knob.knob_margin)/100;

    if (*inner_diameter < 1) *inner_diameter = 1;
}



/************************************************************************
 *
 * CreateSegments
 * 	This knob class procedure is used to create the line segments used
 * in drawing the knob.
 *
 ************************************************************************/
static void CreateSegments( Widget w )
{
    XsmKnobWidget kw = (XsmKnobWidget) w;
    XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);
    unsigned int diameter, inner_diameter;
    unsigned int radius, inner_radius;
    double    angle, cosine, sine, angle_change;
    Position center_x, center_y;
    int       i;
    XSegment   *ptr;
    
    ptr = kw->knob.segments;
    
    angle_change = RADIANS(MAX_ANGLE) /(float)(kw->knob.num_segments - 1);

    (*kc->knob_class.get_diameters)(w, &diameter, &inner_diameter);

    angle = kw->knob.angle_offset;

    center_x = (diameter >> 1) + kw->knob.margin_width +
	       kw->primitive.highlight_thickness;
    center_y = (diameter >> 1) + kw->knob.margin_height +
	       kw->primitive.highlight_thickness; 

    radius = ((diameter >> 1) * 80)/100;

    if (inner_diameter > 2)
       inner_radius = ((inner_diameter >> 1) * 110)/100;
    else
       inner_radius = 1;

    for (i = 0; i < kw->knob.num_segments; i++) {
       cosine = cos(angle);
       sine   = sin(angle);
       ptr->x1   = (short)(center_x + radius * sine);
       ptr->y1 = (short)(center_y - radius * cosine);
       ptr->x2   = (short)(center_x + inner_radius * sine);
       ptr++->y2 = (short)(center_y - inner_radius * cosine);
       angle += angle_change;
    }
}



/************************************************************************
 *
 * DrawIndicator
 * 	This knob class procedure is used to draw the indicator portion of
 * the knob.
 *
 ************************************************************************/
static void DrawIndicator( Widget w )
{
   XsmKnobWidget kw = (XsmKnobWidget) w;
   XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);
   unsigned int diameter, inner_diameter;
   unsigned int radius;
   double angle;
   Position center_x, center_y;

   (*kc->knob_class.get_diameters)(w, &diameter, &inner_diameter);

   angle = (RADIANS(MAX_ANGLE) * kw->knob.value)/ 
		(kw->knob.max_val - kw->knob.min_val) ;

   center_x = (diameter >> 1) + kw->knob.margin_width +
	       kw->primitive.highlight_thickness;
   center_y = (diameter >> 1) + kw->knob.margin_height +
	       kw->primitive.highlight_thickness;

   radius = ((diameter >> 1) * 93)/100;

   diameter = (diameter * 5)/100;
   if (diameter == 0) diameter = 1;

   center_x += (Position)(radius * sin(angle)) - (diameter >> 1);
   center_y -= (Position)(radius * cos(angle)) + (diameter >> 1);


   XDrawArc(XtDisplay(w), XtWindow(w), kw->knob.gc, 
            center_x, center_y, diameter, diameter, 0, 64*360);

   XFillArc(XtDisplay(w), XtWindow(w), kw->knob.gc,
	    center_x, center_y, diameter, diameter, 0, 64*360);

}


/************************************************************************
 *
 * TurnKnob
 * 	This knob class procedure is used in the timeout procedures
 * to show the knob turning.
 *
 ************************************************************************/
/* ARGSUSED */
static void TurnKnob( XtPointer closure, XtIntervalId *id )
{
   Widget w = (Widget) closure;
   XsmKnobWidget kw = (XsmKnobWidget) w;
   XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);
   XsmKnobCallbackStruct cb;

   if ((kw->knob.value < kw->knob.max_val - 1 && kw->knob.move_clockwise) ||
       (kw->knob.value > 1 && !kw->knob.move_clockwise)) {
       kw->knob.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                           (unsigned long) kw->knob.turn_delay,
                                           kc->knob_class.turn, (XtPointer) kw);
   } else {
       kw->knob.turning = False;
   }

   XSetForeground(XtDisplay(kw), kw->knob.gc, kw->knob.knob_color);
   (*kc->knob_class.draw_indicator)(w);

   XDrawSegments(XtDisplay(w), XtWindow(w), kw->knob.gc,
		 kw->knob.segments, kw->knob.num_segments);

   if (kw->knob.move_clockwise) kw->knob.value++;
   else kw->knob.value--;

   if (kw->knob.value_changed_callback) {
      cb.reason = XsmCR_VALUE_CHANGED;
      cb.event = NULL;
      cb.value = kw->knob.value;
      XtCallCallbackList(w, kw->knob.value_changed_callback, (XtPointer) &cb);
   }

   kw->knob.angle_offset = (RADIANS(MAX_ANGLE) * kw->knob.value)/ 
 	                   (kw->knob.max_val - kw->knob.min_val) ;
   (*kc->knob_class.create_segments)(w);

   XSetForeground(XtDisplay(w), kw->knob.gc, kw->knob.indicator_color);
   (*kc->knob_class.draw_indicator)(w);

   XSetForeground(XtDisplay(w), kw->knob.gc, kw->primitive.foreground);
   XDrawSegments(XtDisplay(w), XtWindow(w), kw->knob.gc,
		 kw->knob.segments, kw->knob.num_segments);

}


/************************************************************************
 *
 * DrawKnob
 * 	This knob class procedure is used to draw the knob.
 *
 ************************************************************************/
static void DrawKnob( Widget w )
{
   XsmKnobWidget kw = (XsmKnobWidget) w;
   XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);
   unsigned int diameter, inner_diameter, top_knob_start;

   (*kc->knob_class.get_diameters)(w, &diameter, &inner_diameter);
    
   /* bottom of knob */ 
   XSetForeground(XtDisplay(kw), kw->knob.gc, kw->knob.knob_color);
   XFillArc(XtDisplay(kw), XtWindow(kw), kw->knob.gc,
	    (kw->knob.margin_width + kw->primitive.highlight_thickness), 
	    (kw->knob.margin_height + kw->primitive.highlight_thickness), 
	    diameter, diameter, 0, 64*360);
   XSetForeground(XtDisplay(kw), kw->knob.gc, kw->primitive.foreground);

   XDrawArc(XtDisplay(kw), XtWindow(kw), kw->knob.gc, 
	    (kw->knob.margin_width + kw->primitive.highlight_thickness), 
	    (kw->knob.margin_height + kw->primitive.highlight_thickness), 
	    diameter, diameter, 0, 64*360);

   top_knob_start = (diameter - ((diameter * kw->knob.knob_margin)/100))/2;

   /* top of knob */ 
   XDrawArc(XtDisplay(kw), XtWindow(kw), kw->knob.gc, (top_knob_start +
	    kw->knob.margin_width + kw->primitive.highlight_thickness), 
	    (top_knob_start + kw->knob.margin_height +
	     kw->primitive.highlight_thickness), 
            inner_diameter, inner_diameter, 0, 64*360);

   XDrawSegments(XtDisplay(w), XtWindow(w), kw->knob.gc,
		 kw->knob.segments, kw->knob.num_segments);

   XSetForeground(XtDisplay(w), kw->knob.gc, kw->knob.indicator_color);
   (*kc->knob_class.draw_indicator)(w);
   XSetForeground(XtDisplay(w), kw->knob.gc, kw->primitive.foreground);

}

/*********************************************************************
 *
 * Callproc procedures defined to set default settings for resources
 *
 *********************************************************************/

/******************************************************************
 *  DefaultKnobColor
 *       Get the default knob color Motif default background color.
 ******************************************************************/
static void
DefaultKnobColor( Widget g, int offset, XrmValue *value )
{
    static Pixel            pixel;

    value->addr = (XtPointer) &pixel;
    value->size = sizeof (Pixel);

   /* Work around for bug in _XmBackgroundColorDefault callproc */
    _XmSetDefaultBackgroundColorSpec(XtScreen(g), XmDEFAULT_BACKGROUND);

   /* Motif callprocf to get default color */
    _XmBackgroundColorDefault (g, offset, value);
}

/****************************************************************
 *
 * Action Procedures
 *
 ****************************************************************/


/************************************************************************
 *
 * TurnLeft
 *	An action procedure that is called on a left mouse button press.
 *	This procedure will use a timeout to start the knob from turning
 *	to the left.
 *
 ************************************************************************/
/* ARGSUSED */
static void TurnLeft( Widget w, XEvent *event, char **params,
                      Cardinal *num_params )
{
    XsmKnobWidget kw = (XsmKnobWidget) w;
    XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);
    
   /*
    * As long as the value has not reached the minimum value,
    * continue turning
    */
    if (kw->knob.value > kw->knob.min_val) {
       kw->knob.move_clockwise = False;
       kw->knob.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                       (unsigned long) kw->knob.turn_delay,
                                       kc->knob_class.turn, (XtPointer) w);
    }
}



/* ARGSUSED */
/************************************************************************
 *
 * TurnRight
 *	An action procedure that is called on a right mouse button press.
 *	This procedure will use a timeout to start the knob from turning
 *	to the right.
 *
 ************************************************************************/
static void TurnRight( Widget w, XEvent *event, char **params,
                       Cardinal *num_params )
{
    XsmKnobWidget kw = (XsmKnobWidget) w;
    XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);

   /* As long as the value has not maxed out, continue turning */
    if (kw->knob.value < kw->knob.max_val) {
       kw->knob.move_clockwise = True;
       kw->knob.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                       (unsigned long) kw->knob.turn_delay,
                                       kc->knob_class.turn, (XtPointer) w);
    }
}


/* ARGSUSED */
/************************************************************************
 *
 * ToggleLeft
 *	An action procedure that is called on a key press.
 *	This procedure will add a timeout to start the knob turning left.
 *
 ************************************************************************/
static void ToggleLeft( Widget w, XEvent *event, char **params,
                        Cardinal *num_params )
{
    XsmKnobWidget kw = (XsmKnobWidget) w;
    XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);
    
   /* If the knob has not started turning, added the timeout */
    if (!kw->knob.turning) {
       if (kw->knob.value > kw->knob.min_val) {
          kw->knob.move_clockwise = False;
          kw->knob.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kw->knob.turn_delay,
                                          kc->knob_class.turn, (XtPointer) w);
          kw->knob.turning = True;
       }
    } else {
      /*
       * If it is already turning, remove the old timeout and
       * continue turning with a new timeout
       */
       if (kw->knob.timer_id) XtRemoveTimeOut(kw->knob.timer_id);

      /* if it was turning left continue the turning, otherwise stop. */
       if (kw->knob.move_clockwise) {
          kw->knob.move_clockwise = False;
          kw->knob.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kw->knob.turn_delay,
                                          kc->knob_class.turn, (XtPointer) w);
       } else {
          kw->knob.turning = False;
       }
    }
}


/************************************************************************
 *
 * ToggleRight
 *	An action procedure that is called on a key press.
 *	This procedure will add a timeout to start the knob turning right.
 *
 ************************************************************************/
/* ARGSUSED */
static void ToggleRight( Widget w, XEvent *event, char **params,
                         Cardinal *num_params )
{
    XsmKnobWidget kw = (XsmKnobWidget) w;
    XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);
    
   /* If the knob has not started turning, added the timeout */
    if (!kw->knob.turning) {
       if (kw->knob.value < kw->knob.max_val) {
          kw->knob.move_clockwise = True;
          kw->knob.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kw->knob.turn_delay,
                                          kc->knob_class.turn, (XtPointer) w);
	  kw->knob.turning = True;
       }
    } else {
      /*
       * If it is already turning, remove the old timeout and
       * continue turning with a new timeout
       */
       if (kw->knob.timer_id) XtRemoveTimeOut(kw->knob.timer_id);

      /* if it was turning right continue the turning, otherwise stop. */
       if (!kw->knob.move_clockwise) {
          kw->knob.move_clockwise = True;
          kw->knob.timer_id = XtAppAddTimeOut(XtWidgetToApplicationContext(w),
                                          (unsigned long) kw->knob.turn_delay,
                                          kc->knob_class.turn, (XtPointer) w);
       } else {
          kw->knob.turning = False;
       }
    }
}



/************************************************************************
 *
 * ReleaseKnob
 *	An action procedure that is called on a mouse button release.
 *	This procedure will remove a timeout to stop the knob from turning.
 *
 ************************************************************************/
/* ARGSUSED */
static void ReleaseKnob( Widget w, XEvent *event, char **params,
                         Cardinal *num_params )
{
    XsmKnobWidget kw = (XsmKnobWidget) w;

   /* remove the timeout */
    if (kw->knob.timer_id) XtRemoveTimeOut(kw->knob.timer_id);

    kw->knob.turning = False;
}



/****************************************************************
 *
 * Public Convenience Functions
 *
 ****************************************************************/

/********************************************************
 *
 * XsmCreatWidget() -
 *	Knob widget creation convienence routine.
 *
 ********************************************************/

Widget XsmCreateKnob( Widget parent, char *name, ArgList arglist,
                      Cardinal argcount )
{
    return (XtCreateWidget(name, xsmKnobWidgetClass,
                           parent, arglist, argcount));
}

/**********************************************************************
 *
 * XsmGetKnobValue() -
 *	A function to quickly get the knob value.  If a resource value
 *	needs to be retrieved frequently, it might be a good idea to
 *	have a convenience function like this to avoid costly calls to
 *	XtGetValues().  Too many of these functions can be confusing
 *	the application developer.
 *
 ***********************************************************************/
int XsmGetKnobValue( Widget w )
{
    XsmKnobWidget kw = (XsmKnobWidget) w;

    return(kw->knob.value);
}


/***********************************************************************
 *
 * XsmSetKnobValue() -
 *	A function to quickly set the knob value.  If a resource value
 *	needs to be changed frequently, it might be a good idea to
 *	have a convenience function like this to avoid costly calls to
 *	XtSetValues().  Too many of these functions can be confusing
 *	the application developer.
 *
 ***********************************************************************/
void XsmSetKnobValue( Widget w, int value )
{
   XsmKnobWidget kw = (XsmKnobWidget) w;
   XsmKnobWidgetClass kc = (XsmKnobWidgetClass) XtClass(w);

  /* return if the value doesn't change */

   if (value == kw->knob.value) return;

  /* verify the changed values */

   if (value > kw->knob.max_val) {
      XtWarning("Value is less than MaxValue, Defaulting to MaxValue");
      kw->knob.value = kw->knob.max_val;
   }

   if (value < kw->knob.min_val) {
      XtWarning("Value is greater than MinValue, Defaulting to MinValue");
      kw->knob.value = kw->knob.min_val;
   }

  /*
   * If the knob is realized, redraw the knob segments and indicator to
   * indicate the new value
   */

   if (XtIsRealized(w)) {
     /* erase old indicator */
      XSetForeground(XtDisplay(kw), kw->knob.gc, kw->knob.knob_color);
      (*kc->knob_class.draw_indicator)(w);
      XDrawSegments(XtDisplay(w), XtWindow(w), kw->knob.gc,
		    kw->knob.segments, kw->knob.num_segments);

     /* set new value */
      kw->knob.value = value;
   
     /* recreate segments */
      kw->knob.angle_offset = (RADIANS(MAX_ANGLE) * kw->knob.value)/ 
 	                      (kw->knob.max_val - kw->knob.min_val) ;
      (*kc->knob_class.create_segments)(w);

     /* redraw indicator */
      XSetForeground(XtDisplay(w), kw->knob.gc, kw->knob.indicator_color);
      (*kc->knob_class.draw_indicator)(w);

     /* redraw segments */
      XSetForeground(XtDisplay(w), kw->knob.gc, kw->primitive.foreground);
      XDrawSegments(XtDisplay(w), XtWindow(w), kw->knob.gc,
		    kw->knob.segments, kw->knob.num_segments);
   } else {
      kw->knob.value = value;
   }

}

